""" DataFetcher classes to replace unwieldy dicts """

import argparse
import logging
import sys
from pathlib import Path
from typing import List

# pylint: disable=C0303, E0401, E0611, R0801, R0902, R0903, W0613

from datafetcher.locations_report import (
    LocationsReportBuilder,
)
from datafetcher.retrieve_launcher import ParallelRetrieverLauncher
from datafetcher.utilities import LOG_FORMAT, CLIParam, RetrievalMode

_PROLOGUE = """Retrieve a product (a science product or an ancillary product) 
from the NRAO archive, either by specifying the product's locator or by 
providing the path to a product locator report."""

logger = logging.getLogger(__name__)
formatter = logging.Formatter(LOG_FORMAT)
handler = logging.StreamHandler()
handler.setFormatter(fmt=formatter)
logger.addHandler(handler)


class DataFetcher:
    """
    Given a product locator -or- a .json locations report file,
    pulls indicated product(s) out of NGAS. Fetches into current directory.
    CAPO_PROFILE env var must be set.

    """

    def __init__(self, args_in: List[str]):
        # parse the arguments
        args = self.get_arg_parser().parse_args(args_in)

        # handle the locations report
        if args.location_file is not None:
            self.locations_report = LocationsReportBuilder(Path(args.location_file)).build()
        else:
            self.locations_report = LocationsReportBuilder(args.product_locator).build()

        self.output_dir = Path.cwd()

        # optional arguments
        self.dry_run = args.dry_run
        self.force_overwrite = args.force
        self.direct_copy = args.direct_copy
        self.streaming = args.streaming

        if self.direct_copy or self.streaming:
            # trying to force direct-copy or streaming, rather than letting
            # location report decide
            self.locations_report = self.override_locations_report()

    def override_locations_report(self):
        if self.direct_copy:
            retrieve_method = RetrievalMode.COPY
        elif self.streaming:
            retrieve_method = RetrievalMode.STREAM
        else:
            return self.locations_report

        for target in self.locations_report.targets:
            target.retrieve_method = retrieve_method

        return self.locations_report

    def run(self) -> int:
        """
        So fetch the data already!

        :return:
        """
        launcher = ParallelRetrieverLauncher(
            self.locations_report.targets, self.dry_run, self.force_overwrite
        )
        retrieved_count = launcher.launch_retrievers()
        return retrieved_count

    @staticmethod
    def get_arg_parser() -> argparse.ArgumentParser:
        """Build and return an argument parser with the command line options
            for data-fetcher. This is static because Sphinx
            needs it to build the docs.

            Note difference from previous implementations: Capo profile,
            when needed, will be obtained from the environment;
            files will be retrieved to the current directory.

            Normally, whether to use the NGAS direct-copy plugin to retrieve
            a file is decided on the basis of a file's execution site,
            location, and cluster; the --streaming and --direct-copy flag can
            force one method or the other. If direct-copy is specified but
            not possible, an error will be thrown.

        :return: an argparse 'parser' with command line options for datafetcher.
        """

        parser = argparse.ArgumentParser(
            description=_PROLOGUE,
            formatter_class=argparse.RawTextHelpFormatter,
        )

        exclusive_group = parser.add_mutually_exclusive_group(required=True)

        exclusive_group.add_argument(
            CLIParam.SPL.value,
            action="store",
            dest="product_locator",
            help="product locator to download",
        )
        exclusive_group.add_argument(
            CLIParam.FILE.value,
            action="store",
            dest="location_file",
            help="product locator report (in JSON)",
        )

        dry_or_force_options = parser.add_mutually_exclusive_group(required=False)
        dry_or_force_options.add_argument(
            CLIParam.DRY.value,
            action="store_true",
            dest="dry_run",
            help="dry run; do not fetch product",
            default=False,
        )

        dry_or_force_options.add_argument(
            CLIParam.FORCE.value,
            action="store_true",
            dest="force",
            help="overwrite existing file(s) at dest",
            default=False,
        )

        copy_or_stream_options = parser.add_mutually_exclusive_group(required=False)
        copy_or_stream_options.add_argument(
            CLIParam.DIRECT_COPY.value,
            action="store_true",
            dest="direct_copy",
            help="force direct-copy from NGAS",
            default=False,
        )
        copy_or_stream_options.add_argument(
            CLIParam.STREAMING.value,
            action="store_true",
            dest="streaming",
            help="force streaming from NGAS",
            default=False,
        )

        return parser


def main():
    """
    This is executed when fetcher is launched from the command line.
    :return:
    """

    logging.basicConfig(level=logging.DEBUG)

    args = sys.argv[1:]
    datafetcher = DataFetcher(args)
    datafetcher.run()


if __name__ == "__main__":
    main()
