""" Issues file retrieval requests to NGAS """

# pylint: disable=E0401, E0402, R0801, R0903, R0914, W1202, W1203

import http
import logging
import os
from pathlib import Path
from typing import List
from unittest.mock import patch

import requests
from bs4 import BeautifulSoup
from requests.exceptions import InvalidSchema

from .errors import (
    NGASServiceErrorException,
    SizeMismatchException,
    MissingSettingsException,
    FileErrorException,
    NGASFetchError,
)
from .utilities import RetrievalMode, LOG_FORMAT

from .locations_report import NgasTarget, NgasFile

logger = logging.getLogger(__name__)
formatter = logging.Formatter(LOG_FORMAT)
handler = logging.StreamHandler()
handler.setFormatter(fmt=formatter)
logger.addHandler(handler)

MAX_RETRIES = 5
SLEEP_INTERVAL = 2


class NgasRetriever:
    """ Retrieves one or more files from a specified NGAS server """

    def __init__(self, targets: List[NgasTarget], force_overwrite: bool):
        self.targets = targets
        self.force_overwrite = force_overwrite
        self.attempts_count = 0

    def retrieve(self) -> int:
        """
        Get each file from NGAS, via streaming or direct-copy as indicated

        :return: number of attempts
        """

        for target in self.targets:
            logger.debug(f">>> retrieve method: {target.retrieve_method}")

            destination = build_destination(target.ngas_file)
            if destination.exists() and not self.force_overwrite:
                raise FileExistsError(f"{destination} exists; aborting")
            make_basedir(destination)

            requester = (
                NgasCopier() if target.retrieve_method == RetrievalMode.COPY else NgasStreamer()
            )

            num_tries = 0
            while num_tries < MAX_RETRIES:
                num_tries += 1

                try:
                    requester.send_request(destination, target.ngas_file)
                    self.attempts_count = num_tries
                    return num_tries
                except Exception as exc:
                    if num_tries < MAX_RETRIES:
                        continue
                    raise NGASServiceErrorException from exc


class NgasRequester:
    """ Base class for streaming or direct-copying from NGAS """

    def send_request(self, destination: Path, ngas_file: NgasFile) -> Path:
        """
        Send the stream/direct-copy request to NGAS server

        :param destination: where to put the file
        :param ngas_file: the file to retrieve
        :return:
        """


class NgasStreamer(NgasRequester):
    """ Stream a file from NGAS """

    def send_request(self, destination: Path, ngas_file: NgasFile) -> Path:
        download_url = "http://" + ngas_file.server.server + "/RETRIEVE"
        logger.debug(
            f"attempting streaming download:\nurl: {download_url}\ndestination: {destination}"
        )

        params = {
            "file_id": ngas_file.ngas_file_id,
            "file_version": ngas_file.version,
        }

        try:
            with requests.Session() as session:
                response = session.get(download_url, params=params, stream=True)
                with open(destination, "wb") as file_to_write:
                    for chunk in response.iter_content(chunk_size=8192):
                        file_to_write.write(chunk)
                expected_size = ngas_file.size
                actual_size = os.path.getsize(destination)
                if actual_size == 0:
                    raise FileErrorException(f"{Path(destination).name} was not retrieved")
                if actual_size != expected_size:
                    raise SizeMismatchException(
                        f"expected {Path(destination).name} "
                        f"to be {expected_size} bytes; "
                        f"was {actual_size} bytes"
                    )
        except requests.exceptions.ConnectionError as c_err:
            raise NGASServiceErrorException(
                f"problem connecting with {ngas_file.server.server}"
            ) from c_err
        except AttributeError as a_err:
            logger.warning(f"possible problem streaming: {a_err}")

        if response.status_code != http.HTTPStatus.OK:
            logger.error(f"NGAS does not like this request:\n{response.url}")
            soup = BeautifulSoup(response.text, "lxml-xml")
            ngams_status = soup.NgamsStatus.Status
            message = ngams_status.get("Message")

            raise NGASServiceErrorException(
                {
                    "status_code": response.status_code,
                    "url": response.url,
                    "reason": response.reason,
                    "message": message,
                }
            )

        if check_result(ngas_file, destination):
            logger.debug(f"retrieved {destination} from {ngas_file.server.server}")
            return destination

        raise NGASFetchError(str(destination))


class DryStreamer(NgasStreamer):
    """ Faux NGAS streamer for dry runs and tests """

    def send_request(self, destination: Path, ngas_file: NgasFile) -> Path:
        """
        Here we hit the archiveService but do not write the files.

        :return:
        """
        download_url = "http://" + ngas_file.server.server + "/RETRIEVE"
        logger.debug(
            f"simulating streaming download:\nurl: {download_url}\ndestination: {destination}"
        )

        params = {
            "file_id": ngas_file.ngas_file_id,
            "file_version": ngas_file.version,
        }

        try:
            with requests.Session() as session:
                response = session.get(download_url, params=params, stream=True)
        except requests.exceptions.ConnectionError as c_err:
            raise NGASServiceErrorException(
                f"problem connecting with {ngas_file.server.server}"
            ) from c_err
        except AttributeError as a_err:
            logger.warning(f"possible problem streaming: {a_err}")

        if response.status_code != http.HTTPStatus.OK:
            logger.error(f"NGAS does not like this request:\n{response.url}")
            soup = BeautifulSoup(response.text, "lxml-xml")
            ngams_status = soup.NgamsStatus.Status
            message = ngams_status.get("Message")

            raise NGASServiceErrorException(
                {
                    "status_code": response.status_code,
                    "url": response.url,
                    "reason": response.reason,
                    "message": message,
                }
            )

        return destination


class NgasCopier(NgasRequester):
    """ Retrieves a file from NGAS using the direct-copy plugin """

    def send_request(self, destination: Path, ngas_file: NgasFile) -> Path:
        logger.debug(
            "simulating direct-copy download:\nurl: "
            f"{ngas_file.server.server}\ndestination: {destination}"
        )

        params = {
            "file_id": ngas_file.ngas_file_id,
            "processing": "ngamsDirectCopyDppi",
            "processingPars": "outfile=" + str(destination),
            "file_version": ngas_file.version,
        }
        with requests.Session() as session:
            response = session.get(ngas_file.server.server, params=params)

            if response.status_code != http.HTTPStatus.OK:
                logger.error(f"NGAS does not like this request:\n{response.url}")
                soup = BeautifulSoup(response.text, features="lxml")
                ngams_status = soup.NgamsStatus.Status
                message = ngams_status.get("Message")

                raise NGASServiceErrorException(
                    {
                        "status_code": response.status_code,
                        "url": response.url,
                        "reason": response.reason,
                        "message": message,
                    }
                )


class DryCopier(NgasCopier):
    """ Faux NGAS direct-copy retriever for dry runs and tests """

    def send_request(self, destination: Path, ngas_file: NgasFile) -> Path:
        logger.debug(
            f"simulating direct-copy download:\nurl: "
            f"{ngas_file.server.server}\ndestination: {destination}"
        )

        params = {
            "file_id": ngas_file.ngas_file_id,
            "processing": "ngamsDirectCopyDppi",
            "processingPars": "outfile=" + str(destination),
            "file_version": ngas_file.version,
        }

        with patch("requests.sessions.Session.get"):
            with requests.Session() as session:
                response = session.get(ngas_file.server.server, params=params)

                if response.status_code != http.HTTPStatus.OK:
                    logger.error(f"NGAS does not like this request:\n{response.url}")
                    soup = BeautifulSoup(response.text, features="lxml")
                    ngams_status = soup.NgamsStatus.Status
                    message = ngams_status.get("Message")

                    raise NGASServiceErrorException(
                        {
                            "status_code": response.status_code,
                            "url": response.url,
                            "reason": response.reason,
                            "message": message,
                        }
                    )

        return destination


class NgasNonRetriever(NgasRetriever):
    """ This is the one we'll use for dry runs. """

    def retrieve(self) -> int:
        """
        Get each file from NGAS, via streaming or direct-copy as indicated

        :return: number of attempts
        """
        for target in self.targets:
            logger.debug(f">>> retrieve method: {target.retrieve_method}")

            destination = build_destination(target.ngas_file)
            if destination.exists() and not self.force_overwrite:
                raise FileExistsError(f"{destination} exists; aborting")
            make_basedir(destination)

            requester = (
                DryCopier() if target.retrieve_method == RetrievalMode.COPY else DryStreamer()
            )
            ngas_file = target.ngas_file

            num_tries = 0
            while num_tries < MAX_RETRIES:
                num_tries += 1

                try:
                    requester.send_request(destination, ngas_file)
                    self.attempts_count = num_tries
                    return num_tries
                except Exception as exc:
                    if num_tries < MAX_RETRIES:
                        continue
                    raise exc
                #
                # if target.retrieve_method == RetrievalMode.COPY:
                #     try:
                #         copier = NgasCopier(destination, target.ngas_file)
                #         copier.copy()
                #         self.attempts_count = num_tries
                #         return num_tries
                #     except Exception as exc:
                #         if num_tries < MAX_RETRIES:
                #             continue
                #         raise exc
                # else:
                #     try:
                #         streamer = NgasStreamer(destination, target.ngas_file)
                #         streamer.stream()
                #         self.attempts_count = num_tries
                #         return num_tries
                #     except Exception as exc:
                #         if num_tries < MAX_RETRIES:
                #             continue
                #         raise exc


def build_destination(ngas_file: NgasFile) -> Path:
    """
    Build a destination for the file based on the output directory
    the user provided and the (optional) subdirectory and relative
    path in the file_spec.

    :param ngas_file: product to retrieve from NGAS
    :return: full path to the output file
    """
    output_dir = Path.cwd()
    try:
        if ngas_file.subdirectory is None:
            return output_dir / ngas_file.relative_path
        return output_dir / ngas_file.subdirectory / ngas_file.relative_path
    except (KeyError, TypeError) as err:
        raise MissingSettingsException(err) from err


def check_result(ngas_file: NgasFile, destination: Path) -> bool:
    """
    Confirm that the file was retrieved and its size matches what
    we expect. If not, throw an error and die.

    :param destination: the path to the file to check
    """

    logger.debug(f"verifying fetch of {destination}")
    if not destination.exists():
        raise NGASServiceErrorException(f"file not delivered to {destination}")
    if ngas_file.size != os.path.getsize(destination):
        raise SizeMismatchException(
            f"file retrieval size mismatch on {destination}: "
            f"expected {ngas_file.size}, "
            f"got {os.path.getsize(destination)}"
        )
    logger.debug("\tlooks good; sizes match")
    return True


def make_basedir(destination: Path) -> Path:
    """
    Creates the directory the product will be saved to, if it doesn't exist.

    :param destination: pathname of file to be retrieved
    :return: the output dir
    """
    basedir = destination.parent
    if basedir.is_dir() and not os.access(basedir, os.W_OK):
        raise FileErrorException(f"output directory {basedir} is not writable")
    try:
        umask = os.umask(0o000)
        basedir.mkdir(parents=True, exist_ok=True)
        os.umask(umask)
        return basedir
    except OSError as os_err:
        raise FileErrorException(f"failure trying to create output directory {basedir}") from os_err
