""" Launches NGAS retrievers in parallel """

import copy
import logging
from typing import List, Dict

# pylint: disable=E0402, E0611, R0801, W1203

from .ngas_retriever import NgasRetriever, NgasNonRetriever
from .utilities import LOG_FORMAT

from .locations_report import NgasTarget

logger = logging.getLogger(__name__)
formatter = logging.Formatter(LOG_FORMAT)
handler = logging.StreamHandler()
handler.setFormatter(fmt=formatter)
logger.addHandler(handler)


class ParallelRetrieverLauncher:
    """Replacement for Ye Olde ParallelFetcher.
    We're no longer passing in the output dir, because now it's just the
    current directory
    """

    def __init__(self, ngas_targets: List[NgasTarget], dry_run: bool, force_overwrite: bool):
        self.ngas_targets = ngas_targets
        self.files_by_server = self.organize_file_info()
        self.dry_run = dry_run
        self.force_overwrite = force_overwrite

    def organize_file_info(self) -> Dict:
        """The location report we get back looks like this, for each file:

        {"ngas_file_id":"17B-197_2018_02_19_T15_59_16.097.tar",
        "subdirectory":"17B-197.sb34812522.eb35115211.58168.58572621528",
        "relative_path":"17B-197_2018_02_19_T15_59_16.097.tar",
        "checksum":"-1848259250",
        "checksum_type":"ngamsGenCrc32",
        "version":1,
        "size":108677120,
        "server":{"server":"nmngas01.aoc.nrao.edu:7777",
            "location":"DSOC",
            "cluster":"DSOC"
        }}

        Re-organize it to group files under servers so we can get all the
        files from each server in parallel.
        """

        servers = set([ngas_target.ngas_file.server for ngas_target in self.ngas_targets])

        result = {}
        for server in servers:
            targets = []
            for ngas_target in self.ngas_targets:
                target_copy = copy.deepcopy(ngas_target)
                if target_copy.ngas_file.server == server:
                    targets.append(target_copy)
            result[server] = targets

        return result

    def launch_retrievers(self) -> int:
        """
        Launch NGAS retrievers in parallel.

        :return: number of files retrieved
        """

        if self.dry_run:
            retriever = NgasNonRetriever(self.ngas_targets, self.force_overwrite)
        else:
            retriever = NgasRetriever(self.ngas_targets, self.force_overwrite)

        num_files = 0
        count = 0
        for server_url, targets in self.files_by_server.items():
            logger.debug(f"retrieving {len(targets)} from {server_url}....")
            num_files += len(targets)
            for target in targets:
                count += 1
                logger.debug(
                    f">>> retrieving {target.relative_path} "
                    f"({target.size} bytes, "
                    f"no. {count} of {num_files}) via "
                    f"{target.retrieve_method}...."
                )

                retriever.retrieve()

        return count
