"""
A Quasar is a radio emitting active galactic nucleus (i.e. an actively feeding supermassive black hole).
Radio Astronomy for the win!!!!
"""
import argparse
import logging
import os
import sys

import glob
import re

from pycapo import CapoConfig

from .interfaces import CasaEmulatorIF
from .forger import forge

"""
The Vela system allows for testing Workspaces workflows without submitting to CASA on the cluster.
"""

logger = logging.getLogger("vela")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


class CalibrationEmulator(CasaEmulatorIF):
    def __init__(self, metadata: str,  ppr: str):
        self.ppr = ppr
        self.metadata = metadata

    def run(self):
        logger.info("Vela waiting on the Forger....")
        forge(self.metadata)

    def setup_environment(self, parameters: dict):
        os.environ['SCIPIPE_ROOTDIR'] = parameters["rootDirectory"]
        os.environ['CASA_HOME'] = parameters["homeForReprocessing"]
        os.environ['PPR_FILENAME'] = str(self.ppr)

        logger.info("Checking processing environment:")
        env1 = os.environ.get('SCIPIPE_ROOTDIR')
        logger.info(f"SCIPIPE_ROOTDIR: {env1}")
        env2 = os.environ.get('CASA_HOME')
        logger.info(f"CASA_HOME: {env2}")
        env3 = os.environ.get('PPR_FILENAME')
        logger.info(f"PPR_FILENAME: {env3}")
        if 'None' in [env1, env2, env3]:
            logger.info("Environment setup Failed!")
        else:
            logger.info("Environment ready for processing")

    def check_logs(self):
        logger.info("CHECKING VELA LOGS!")

        casa_logs = glob.glob('vela-*.log')

        for file in casa_logs:
            if re.match("^.*SEVERE\sflagmanager.*$", open(file).read()):
                logger.info("VELA ERROR! Please inspect logs.")
            else:
                logger.info("VELA Success!")


def _get_settings():
    use_casa = CapoConfig().getboolean("edu.nrao.archive.workspaces.ProcessingSettings.useCasa")
    root_dir = CapoConfig().settings("edu.nrao.archive.workspaces.ProcessingSettings").rootDirectory
    casa_home = CapoConfig().settings("edu.nrao.archive.workflow.config.CasaVersions").homeForReprocessing

    return {"useCasa": use_casa,"homeForReprocessing": casa_home, "rootDirectory": root_dir}


def parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Workspaces VELA processing launcher",
        formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "--standard-cal",
        nargs=2,
        action="store",
        required=False,
        help="run the standard calibration pipeline"
    )
    parser.add_argument(
        "--standard-image",
        nargs=2,
        action="store",
        required=False,
        help="run the standard imaging pipeline"
    )
    return parser


def main():
    args = parser().parse_args()
    metadata = args.standard_cal[0]
    ppr = args.standard_cal[1]

    settings = _get_settings()
    if args.standard_cal:
        emulator = CalibrationEmulator(metadata, ppr)
        emulator.setup_environment(settings)
        emulator.run()
        emulator.check_logs()
    else:
        logger.info("ARGUMENT ERROR: no valid argument was provided.")
